from drepr.version import __version__
from drepr.engine import execute, FileOutput, MemoryOutput, OutputFormat
from drepr.graph_deprecated import Graph
from drepr.models import DRepr, DReprBuilder
from drepr import outputs
from drepr import models


import copy
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Union, Tuple, List

import ujson, traceback

# noinspection PyUnresolvedReferences
from drepr import drepr_engine
from drepr.version import __engine_version__
from drepr.models import DRepr, DEFAULT_RESOURCE_ID, SemanticModel, Alignment, AlignmentType, RangeAlignment, \
    AlignedStep, ValueAlignment, LiteralNode, ClassNode
from drepr.patches import ResourceData, ResourceDataFile, ResourceDataString, xml_patch, jp_propname_patch, nc_patch, \
    static_class_patch

assert drepr_engine.__version__ == __engine_version__, f"You are using a different version of D-REPR" \
                                                          f": {drepr_engine.__version__}. The correct one is" \
                                                          f": {__engine_version__}"


def execute(ds_model: DRepr,
            resources: Union[str, Dict[str, Union[str, ResourceData]]],
            output: "Output",
            debug: bool = False):
    ptr = None
    if isinstance(resources, (str, tuple)):
        resources = {DEFAULT_RESOURCE_ID: resources}

    # normalize resources so that we know which one is from files and which one is from string. Below is the schema
    # resources = {
    #   <resource_id>: { "file"|"string": <value> }
    # }
    resources = {
        rid: ResourceDataFile(resource) if type(resource) is str else resource
        for rid, resource in resources.items()
    }

    ds_model = nc_patch.patch(ds_model, resources)
    ds_model = xml_patch.patch(ds_model, resources)
    ds_model = jp_propname_patch.patch(ds_model, resources)
    ds_model, resources = static_class_patch.patch(ds_model, resources)

    if isinstance(output, FileOutput):
        engine_output = {
            "file": {
                "fpath": str(output.fpath),
                "format": output.format.value
            }
        }
    elif isinstance(output, MemoryOutput):
        engine_output = {
            "memory": {
                "format": output.format.value
            }
        }
    else:
        raise NotImplementedError()

    try:
        engine_model = ds_model.to_engine_format()
        if debug:
            print(">>> the engine is going to execute the below drepr model")
            print(ujson.dumps(engine_model.model, indent=4))

        ptr = drepr_engine.create_executor(ujson.dumps({
            "resources": [
                {"file": resources[rid].file_path} if isinstance(resources[rid], ResourceDataFile) else {"string": resources[rid].value}
                for rid in sorted(resources.keys(), key=lambda k: engine_model.resource_idmap[k])
            ],
            "output": engine_output,
            "edges_optional": engine_model.edges_optional,
            "description": engine_model.model
        }))
        if debug:
            print(f"""
******************************************
>>> Execution plan:
{drepr_engine.get_exec_plan(ptr)}
******************************************
""")

        result = drepr_engine.run_executor(ptr, ujson.dumps(engine_output))
        if isinstance(output, MemoryOutput) and output.format == OutputFormat.GraphPy:
            class2nodes = {}
            for u in ds_model.sm.iter_class_nodes():
                class2nodes[u.node_id] = result['class2nodes'][engine_model.sm_node_idmap[u.node_id]]
            return class2nodes
        return result
    finally:
        if ptr is not None:
            drepr_engine.destroy_executor(ptr)


def complete_description(ds_model: DRepr) -> 'CompleteDescription':
    patched_model = jp_propname_patch.patch(ds_model, None)
    # Note: do not patch the static class because we do post-update
    # description for the static class
    # patched_model, resources = static_class_patch.patch(ds_model, resources)

    engine_model = patched_model.to_engine_format()
    extra_info = drepr_engine.complete_description(ujson.dumps(engine_model.model))

    if ds_model is patched_model:
        sm = copy.deepcopy(ds_model.sm)
    else:
        sm = patched_model.sm

    # update subjects
    subjs = {}
    for n in sm.iter_class_nodes():
        class_id = engine_model.sm_node_idmap[n.node_id]
        aidx = extra_info['class2subj'][class_id]
        if aidx == -1:
            # just need to pick a random literal node
            for e in sm.iter_outgoing_edges(n.node_id):
                if isinstance(sm.nodes[e.target_id], LiteralNode):
                    subjs[n.node_id] = e.target_id
                    # subjs.add((n.node_id, e.target_id))
                    break
        else:
            subjs[n.node_id] = f'dnode:{patched_model.attrs[aidx].id}'
            # subjs.add((n.node_id, f'dnode:{new_ds_model.attrs[aidx].id}'))

    for e in sm.edges.values():
        if subjs.get(e.source_id, None) == e.target_id:
            e.is_subject = True

    # update alignments
    alignments = dict()
    for align in patched_model.aligns:
        alignments[(f"dnode:{align.source}", f"dnode:{align.target}")] = [align]

    for (source, target), aligns_lst in extra_info['aligned_funcs'].items():
        source = f"dnode:{ds_model.attrs[source].id}"
        target = f"dnode:{ds_model.attrs[target].id}"
        aligns = []
        for align in aligns_lst:
            align_type = AlignmentType(align['type'])
            if align_type == AlignmentType.range:
                aligns.append(RangeAlignment(
                    ds_model.attrs[align['source']].id,
                    ds_model.attrs[align['target']].id,
                    [AlignedStep(**o) for o in align['aligned_dims']]
                ))
            elif align_type == AlignmentType.value:
                aligns.append(ValueAlignment(
                    ds_model.attrs[align['source']].id,
                    ds_model.attrs[align['target']].id,
                ))
            else:
                raise NotImplementedError()
        alignments[source, target] = aligns

    for n in sm.iter_class_nodes():
        source = subjs[n.node_id]
        for e in sm.iter_outgoing_edges(n.node_id):
            v = sm.nodes[e.target_id]
            if isinstance(v, LiteralNode) and v.node_id != source:
                alignments[source, e.target_id] = [RangeAlignment(source, e.target_id, [])]
            elif isinstance(v, ClassNode) and (source, subjs[e.target_id]) not in alignments:
                # link to literal class
                assert isinstance(sm.nodes[subjs[e.target_id]], LiteralNode)
                alignments[source, subjs[e.target_id]] = [RangeAlignment(source, subjs[e.target_id], [])]

    return CompleteDescription(sm, alignments)


@dataclass
class CompleteDescription:
    sm: SemanticModel
    alignments: Dict[Tuple[str, str], List[Alignment]]


class OutputFormat(Enum):
    TTL = "ttl"
    GraphJSON = "graph_json"
    GraphPy = "graph_py"


@dataclass
class FileOutput:
    fpath: str
    format: OutputFormat


@dataclass
class MemoryOutput:
    format: OutputFormat


Output = Union[FileOutput, MemoryOutput]



from collections import defaultdict
from dataclasses import asdict
from typing import TYPE_CHECKING, List

from drepr.models.parse_v1.path_parser import PathParserV1
from drepr.utils.validator import *
from .align_parser import AlignParser
from .attr_parser import AttrParser
from .preprocessing_parser import PreprocessingParser
from .resource_parser import ResourceParser
from .sm_parser import SMParser
from ..align import AlignmentType, RangeAlignment
from ..sm import SemanticModel, ClassNode, DataNode, LiteralNode

if TYPE_CHECKING:
    from ..drepr import DRepr


class ReprV1Parser:
    """
    The DREPR language version 1 has the following schema:

    ```
    version: '1'
    resources: <resources>
    [preprocessing]: <preprocessing> (default is empty list)
    attributes: <attributes>
    [alignments]: <alignments> (default is empty list)
    semantic_model: <semantic_model>
    ```
    """

    TOP_KEYWORDS = {
        "version", "resources", "preprocessing", "attributes", "alignments", "semantic_model"
    }
    DEFAULT_RESOURCE_ID = "default"

    @classmethod
    def parse(cls, raw: dict):
        from ..drepr import DRepr

        Validator.must_be_subset(
            cls.TOP_KEYWORDS,
            raw.keys(),
            setname="Keys of D-REPR configuration",
            error_msg="Parsing D-REPR configuration")

        for prop in ['version', 'resources', 'attributes']:
            Validator.must_have(raw, prop, error_msg="Parsing D-REPR configuration")

        Validator.must_equal(raw['version'], '1', "Parsing D-REPR configuration version")
        resources = ResourceParser.parse(raw['resources'])

        if len(resources) == 1:
            default_resource_id = resources[0].id
        else:
            default_resource_id = ResourceParser.DEFAULT_RESOURCE_ID

        path_parser = PathParserV1()
        preprocessing = PreprocessingParser(path_parser).parse(default_resource_id, resources, raw.get('preprocessing', []))
        attrs = AttrParser(path_parser).parse(default_resource_id, resources, raw['attributes'])
        aligns = AlignParser.parse(raw.get('alignments', []))

        if 'semantic_model' in raw:
            sm = SMParser.parse(raw['semantic_model'])
            sm.prefixes.update(SemanticModel.get_default_prefixes())
        else:
            sm = None

        return DRepr(resources, preprocessing, attrs, aligns, sm)

    @classmethod
    def dump(cls, drepr: 'DRepr', simplify: bool = True, use_json_path: bool = False):
        version = '1'
        sm = OrderedDict([("data_nodes", OrderedDict()), ("relations", []), ("literal_nodes", []),
                          ("subjects", OrderedDict([])), ("prefixes", drepr.sm.prefixes)])

        class_ids: Dict[str, Dict[str, str]] = defaultdict(lambda: {})
        for node in drepr.sm.nodes.values():
            if isinstance(node, ClassNode):
                class_ids[node.label][
                    node.node_id] = f"{node.label}:{len(class_ids[node.label]) + 1}"

        for node in drepr.sm.nodes.values():
            if isinstance(node, DataNode):
                edge = [e for e in drepr.sm.edges.values() if e.target_id == node.node_id][0]
                sm['data_nodes'][
                    node.
                    attr_id] = f"{class_ids[drepr.sm.nodes[edge.source_id].label][edge.source_id]}--{edge.label}"
                if node.data_type is not None:
                    sm['data_nodes'][node.attr_id] += f"^^{node.data_type.value}"

            if isinstance(node, LiteralNode):
                edge = [e for e in drepr.sm.edges if e.target_id == node.node_id][0]
                sm['literal_nodes'].append(
                    f"{class_ids[drepr.sm.nodes[edge.source_id].label][edge.source_id]}--{edge.label}--{node.value}"
                )
                if node.data_type is not None:
                    sm['literal_nodes'][-1] += f"^^{node.data_type.value}"

        for edge in drepr.sm.edges.values():
            if isinstance(drepr.sm.nodes[edge.source_id], ClassNode) and isinstance(
                    drepr.sm.nodes[edge.target_id], ClassNode):
                sm['relations'].append(
                    f"{class_ids[drepr.sm.nodes[edge.source_id].label][edge.source_id]}--{edge.label}--{class_ids[drepr.sm.nodes[edge.target_id].label][edge.target_id]}"
                )

            if edge.is_subject:
                sm['subjects'][class_ids[drepr.sm.nodes[edge.source_id].label][
                    edge.source_id]] = drepr.sm.nodes[edge.target_id].attr_id

        preprocessing: List[dict] = []
        for prepro in drepr.preprocessing:
            preprocessing.append(OrderedDict([("type", prepro.type.value)]))
            for k, v in asdict(prepro.value).items():
                preprocessing[-1][k] = v
            preprocessing[-1]["path"] = prepro.value.path.to_lang_format(use_json_path)

        return OrderedDict(
            [("version", version),
             ("resources",
              OrderedDict(
                  [(res.id,
                    OrderedDict([("type", res.type.value)] + (
                        [(k, v)
                         for k, v in asdict(res.prop).items()] if res.prop is not None else [])))
                   for res in drepr.resources])), ("preprocessing", preprocessing),
             ("attributes",
              OrderedDict([(attr.id,
                            OrderedDict([("resource_id", attr.resource_id),
                                         ("path", attr.path.to_lang_format(use_json_path)),
                                         ("unique", attr.unique), ("sorted", attr.sorted.value),
                                         ("value_type", attr.value_type.value),
                                         ("missing_values", attr.missing_values)]))
                           for attr in drepr.attrs])),
             ("alignments", [
                 OrderedDict([("type", AlignmentType.range.value), ("source", align.source),
                              ("target", align.target),
                              ("aligned_dims", [
                                  OrderedDict([
                                      ("source", step.source_idx),
                                      ("target", step.target_idx),
                                  ]) for step in align.aligned_steps
                              ])]) if isinstance(align, RangeAlignment) else
                 OrderedDict([("type", AlignmentType.value.value), ("source", align.source),
                              ("target", align.target)]) for align in drepr.aligns
             ]), ("semantic_model", sm)])



from abc import ABC, abstractmethod
from typing import List, TYPE_CHECKING, Optional

from drepr.models import Edge

if TYPE_CHECKING:
    from drepr.outputs.base_lst_output_class import BaseLstOutputClass
from drepr.outputs.prop_data_ndarray import PropDataNDArray


class BaseOutputPredicate(ABC):
    # URI of the predicate
    uri: str
    # list of outgoing edges of the same class that represent the same predicate.
    edges: List[Edge]

    @abstractmethod
    def ndarray_size(self) -> int:
        """Return the number of values of this predicate when it is obtained using the method `as_ndarray([])`"""
        pass

    @abstractmethod
    def as_ndarray(self, indexed_predicates: List['BaseOutputPredicate']) -> PropDataNDArray:
        """Get values of this predicate as nd-array, where each item in the array is a value of this predicate
        of one instance.

        This function will throw exceptions if values of the predicate of one instance are more than one.
        """
        pass

    @abstractmethod
    def o(self) -> Optional['BaseLstOutputClass']:
        """
        Get targets if this predicate is object property.
        We have more than one target when the same predicate is linked to classes that have the same URI.
        """
        pass



from typing import List

from drepr.outputs.base_output_sm import BaseOutputSM


def test_iter_records_s01_2(s01: List[BaseOutputSM], s02: List[BaseOutputSM]):
    # basic access
    for sm in (item for lst in [s01, s02] for item in lst):
        cls = sm.c("mint:Variable")[0]
        records = list(cls.iter_records())
        assert len(records) == 25
        assert len(records) == len({r.id for r in records})
        assert len({r.s("mint:raster") for r in records}) == 1
        r = records[6]
        d = r.to_dict()
        d.pop('@id')
        d.pop('mint:raster')
        assert d == {
            "mint-geo:lat": [20.2],
            "mint-geo:long": [44.2],
            "rdf:value": [-0.43058764]
        }
        assert r.s("mint-geo:lat") == 20.2 and r.m("mint-geo:lat") == [20.2]
        assert r.s("mint-geo:long") == 44.2 and r.m("mint-geo:long") == [44.2]
        assert r.s("rdf:value") == -0.43058764 and r.m("rdf:value") == [-0.43058764]



from typing import List, Dict, Tuple, Callable, Any, Optional, Union

import ujson
import numpy as np
from drepr.executors.readers.ra_reader import NDArrayReader, Index
from drepr.models import IndexExpr, RangeExpr


class NPDictReader(NDArrayReader):
    """
    A key-value resource for the purpose of testing. Any array in this resource will be read as numpy array
    """
    def __init__(self, data):
        self.data = data

    @classmethod
    def from_file(cls, infile: str):
        with open(infile, "r") as f:
            data = ujson.load(f)
            for k, v in data.items():
                if isinstance(v, list):
                    data[k] = np.asarray(v)
        return NPDictReader(data)

    def get_value(self, index: List[Index]):
        return self.data[index[0]]

    def set_value(self, index: List[Index], value):
        raise NotImplementedError()

    def select(self, steps: List[Union[IndexExpr, RangeExpr]]):
        assert isinstance(steps[0], IndexExpr)
        return self.data[steps[0].val]

    def len(self) -> int:
        return len(self.data)

    def len_range(self) -> int:
        raise Exception("Error in your drepr model")


from collections import defaultdict
from dataclasses import dataclass
from typing import Union, Dict, List, Iterable, Optional, Tuple

import ujson

from drepr.models import DRepr
from drepr.engine import execute, MemoryOutput, OutputFormat
from drepr.models.sm import ClassNode


@dataclass
class Edge:
    id: int
    source: int
    target: int
    label: str


@dataclass
class Node:
    id: int
    data: dict
    edges_out: List[int]
    edges_in: List[int]


class Graph:
    def __init__(self, nodes: List[Node], edges: List[Edge], ds_model: Optional[DRepr] = None):
        self.nodes = nodes
        self.edges = edges

        if ds_model is not None:
            self.prefixes: Dict[str, str] = ds_model.sm.prefixes
            self.class2nodes: Dict[str, List[int]] = {
                n.label: []
                for n in ds_model.sm.nodes.values() if isinstance(n, ClassNode)
            }
        else:
            self.prefixes: Dict[str, str] = {}
            self.class2nodes: Dict[str, List[int]] = defaultdict(lambda: [])

        self.class2nodes[None] = []
        for n in nodes:
            self.class2nodes[n.data.get('@type', None)].append(n.id)

    def iter_nodes(self) -> Iterable[Node]:
        return self.nodes

    def iter_nodes_by_class(self, cls: str) -> Iterable[Node]:
        for nid in self.class2nodes[cls]:
            yield self.nodes[nid]

    def iter_edges(self) -> Iterable[Edge]:
        return self.edges

    @staticmethod
    def from_drepr(ds_model: DRepr, resources: Union[str, Dict[str, str]]) -> "Graph":
        result = execute(ds_model, resources, MemoryOutput(OutputFormat.GraphJSON))
        ser_nodes = result["nodes"].split("\n")
        ser_edges = result["edges"].split("\n")
        assert ser_nodes[-1] == "" and ser_edges[-1] == ""

        # remove last elements
        ser_nodes.pop()
        ser_edges.pop()

        nodes, edges = Graph._deserialize_drepr_output(ser_nodes, ser_edges)
        return Graph(nodes, edges, ds_model)

    @staticmethod
    def from_drepr_output_file(node_file: str, edge_file: str) -> "Graph":
        with open(node_file, "r") as f:
            ser_nodes = f.readlines()

        with open(edge_file, "r") as f:
            ser_edges = f.readlines()

        nodes, edges = Graph._deserialize_drepr_output(ser_nodes, ser_edges)
        return Graph(nodes, edges, None)

    def serialize(self, fpath: str):
        with open(fpath, 'w') as f:
            ujson.dump(f, {"prefixes": self.prefixes, "nodes": self.nodes, "edges": self.edges})

    @staticmethod
    def _deserialize_drepr_output(ser_nodes: List[str], ser_edges: List[str]) -> Tuple[List[Node], List[Edge]]:
        nodes = []
        edges = []
        for ser_node in ser_nodes:
            try:
                u = ujson.loads(ser_node)
            except:
                print(ser_node)
                raise
            nodes.append(Node(u['id'], u['data'], [], []))

        for ser_edge in ser_edges:
            sid, tid, lbl = ser_edge.split("\t")
            sid = int(sid)
            tid = int(tid)

            eid = len(edges)
            edges.append(Edge(eid, sid, tid, lbl))
            nodes[tid].edges_in.append(eid)
            nodes[sid].edges_out.append(eid)

        return nodes, edges


from abc import ABC, abstractmethod
from typing import List, Tuple, Any, Iterable, Union

from drepr.outputs.base_output_class import BaseOutputClass
from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_record import BaseRecord


class BaseSubsetOutputClass(BaseOutputClass):
    pass



import copy
from typing import List, Dict, Tuple, Callable, Any, Optional

from drepr.outputs.base_record import BaseRecord


class GraphRecord(BaseRecord):

    def __init__(self, data: dict) -> None:
        super().__init__()
        self.data = data

    @property
    def id(self):
        return self.data['@id']

    def s(self, predicate_uri: str) -> Any:
        try:
            return self.data[predicate_uri][0]
        except KeyError:
            return None

    def m(self, predicate_uri: str) -> Any:
        return self.data[predicate_uri]

    def us(self, predicate_uri: str, val: Any):
        if len(self.data[predicate_uri]) == 0:
            self.data[predicate_uri].append(val)
        else:
            self.data[predicate_uri][0] = val

    def um(self, predicate_uri: str, val: Any):
        self.data[predicate_uri] = val

    def to_dict(self) -> dict:
        return copy.deepcopy(self.data)



from collections import defaultdict
from typing import List, Tuple, Any, Union, Iterable, TYPE_CHECKING, Dict, Optional

from drepr.models import ClassNode, SemanticModel, DRepr
from drepr.outputs.record_id import RecordID, GraphRecordID
from drepr.outputs.base_output_class import BaseOutputClass
from drepr.outputs.base_record import BaseRecord

if TYPE_CHECKING:
    from drepr.outputs.graph_backend.graph_backend import GraphBackend
from drepr.outputs.graph_backend.graph_predicate import GraphPredicate
from drepr.outputs.graph_backend.graph_record import GraphRecord
from drepr.outputs.graph_backend.subset_graph_class import SubsetGraphClass


class GraphClass(BaseOutputClass):

    def __init__(self, backend: 'GraphBackend', u: ClassNode, nodes: List[dict]) -> None:
        super().__init__()
        self.backend = backend
        self.cls = u
        self.nodes = {v['@id']: GraphRecord(v) for v in nodes}
        self.predicates: Dict[str, GraphPredicate] = defaultdict(list)
        self._is_blank = True
        uri2edges = defaultdict(list)
        for e in backend.drepr.sm.iter_outgoing_edges(u.node_id):
            if e.label == 'drepr:uri':
                self._is_blank = False
            uri2edges[e.label].append(e)

        for uri, edges in uri2edges.items():
            self.predicates[uri] = GraphPredicate(backend, self, edges)

    def _init(self, backend: 'GraphBackend'):
        for pred in self.predicates.values():
            pred._init(backend)

    @property
    def id(self):
        return self.cls.node_id

    @property
    def uri(self):
        return self.cls.label

    def is_blank(self) -> bool:
        return self._is_blank

    def iter_records(self) -> Iterable[GraphRecord]:
        return iter(self.nodes.values())

    def get_record_by_id(self, rid: RecordID) -> BaseRecord:
        return self.nodes[rid]

    def p(self, predicate_uri: str) -> Optional[GraphPredicate]:
        return self.predicates.get(predicate_uri, None)

    def o(self, predicate_uri: str, target_uri: str) -> List['GraphClass']:
        return self.index_po[predicate_uri, target_uri]

    def filter(self, conditions: Union['FCondition', List['FCondition']]) -> 'SubsetGraphClass':
        pass

    def group_by(self, predicate: Union[str, GraphPredicate]) -> Iterable[Tuple[Any, 'SubsetGraphClass']]:
        if isinstance(predicate, GraphPredicate):
            predicate = predicate.uri

        groups = {}
        for r in self.nodes.values():
            if r.s(predicate) not in groups:
                groups[r.s(predicate)] = SubsetGraphClass(self, [r])
            else:
                groups[r.s(predicate)].nodes.append(r)

        return iter(groups.items())

    def __len__(self):
        return len(self.nodes)


from typing import List, Tuple, Any, TYPE_CHECKING

from drepr.outputs.record_id import RecordID
from drepr.outputs.base_record import BaseRecord

if TYPE_CHECKING:
    from drepr.outputs.array_backend.array_class import ArrayClass


class ArrayRecord(BaseRecord):
    """
    This record assumes that there is one to one mapping between an item in the pk attribute with the subject attribute.
    If this assumption does not hold, we have to perform a group by the subject attribute to gather duplicated record first.
    """

    def __init__(self, index: Tuple[int, ...], array_class: 'ArrayClass'):
        self._index = index
        self._cls = array_class

    @property
    def id(self) -> RecordID:
        """Get ID of the record"""
        return self._cls.uri_attr.get_sval(self._index)

    def s(self, pred_uri: str):
        """Get value of a predicate, guarantee to return the first single value"""
        aid = self._cls.pred2attrs[pred_uri][0]
        return self._cls.attrs[aid].get_sval(self._index)

    def m(self, pred_uri: str):
        """Get values of a predicate, always return a list"""
        result = []
        for aid in self._cls.pred2attrs[pred_uri]:
            result += self._cls.attrs[aid].get_mval(self._index)
        return result

    def us(self, predicate: str, val: Any):
        aid = self._cls.pred2attrs[predicate][0]
        self._cls.attrs[aid].set_value(self._index, val)

    def um(self, predicate: str, values: List[Any]):
        values = iter(values)
        for i, aid in enumerate(self._cls.pred2attrs[predicate]):
            self._cls.attrs[aid].set_value(self._index, values)

    def to_dict(self) -> dict:
        info = {'@id': self.id}
        for k in self._cls.pred2attrs.keys():
            info[k] = self.m(k)
        return info



import copy
from typing import Dict

import ujson
from netCDF4 import Dataset

from drepr.models import ResourceType, DRepr
from drepr.patches import ResourceData, ResourceDataFile, ResourceDataString


def patch(repr: DRepr, resources: Dict[str, ResourceData]) -> DRepr:
    """
    This patch will turn any netcdf resources to JSON
    """
    need_patch = False
    netcdf_types = {ResourceType.NetCDF4, ResourceType.NetCDF3}
    for resource in repr.resources:
        if resource.type in netcdf_types:
            need_patch = True

    if need_patch:
        repr = copy.deepcopy(repr)
        for resource in repr.resources:
            if resource.type in netcdf_types:
                resource.type = ResourceType.JSON

                assert isinstance(
                    resources[resource.id],
                    ResourceDataFile), "Doesn't support loading netcdf from raw bytes yet"
                nc_data = Dataset(resources[resource.id].file_path,
                                  "r+",
                                  format=ResourceType.NetCDF4.value.upper())

                doc = {}
                for vname in nc_data.variables.keys():
                    doc[vname] = nc_data.variables[vname][:].tolist()

                resources[resource.id] = ResourceDataString(ujson.dumps(doc))
                # resources[resource.id] = ResourceDataFile(f"/tmp/{str(uuid.uuid4())}.json")
                # resources[resource.id] = ResourceDataFile("/tmp/bbfad741-f13f-4227-9a3c-9ab5fbe54bc4.json")
                # with open(resources[resource.id].file_path, "w") as f:
                #     ujson.dump(doc, f)

    return repr



from abc import ABC, abstractmethod
from typing import List, Tuple, Any, Iterable, Union, Optional

from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_record import BaseRecord
from drepr.outputs.record_id import RecordID


class BaseOutputClass(ABC):
    # id of the class in the semantic model
    id: str
    uri: str

    @abstractmethod
    def is_blank(self) -> bool:
        """Return true if the class does not has `drepr:uri` predicate"""
        pass

    @abstractmethod
    def iter_records(self) -> Iterable[BaseRecord]:
        """Iterate through each record of this class"""
        pass

    @abstractmethod
    def get_record_by_id(self, rid: RecordID) -> BaseRecord:
        """Get a record by id"""
        pass

    @abstractmethod
    def p(self, predicate_uri: str) -> Optional[BaseOutputPredicate]:
        """Get a outgoing predicate by URI. Return None if the class does not have URI"""
        pass

    @abstractmethod
    def filter(self, conditions: Union['FCondition', List['FCondition']]) -> 'BaseOutputClass':
        """Filter out records that does not satisfied the giving condition"""
        pass

    @abstractmethod
    def group_by(self, predicate: Union[str, BaseOutputPredicate]) -> Iterable[Tuple[Any, 'BaseOutputClass']]:
        """Group records of this classes by values of a predicate"""
        pass



from typing import List, Dict, Tuple, Callable, Any, Optional, Iterable, TYPE_CHECKING

if TYPE_CHECKING:
    from drepr.outputs.array_backend.array_class import ArrayClass
from drepr.outputs.array_backend.array_record import ArrayRecord
from drepr.outputs.array_backend.subset_array_class import SubsetArrayClass
from drepr.outputs.record_id import RecordID
from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_lst_output_class import BaseLstOutputClass
from drepr.outputs.base_output_class import BaseOutputClass


class LstArrayClass(BaseLstOutputClass):

    def __iter__(self) -> Iterable['ArrayClass']:
        return iter(self.classes)

    def __len__(self) -> int:
        return len(self.classes)

    def __getitem__(self, item: int) -> BaseOutputClass:
        return self.classes[item]

    def p(self, predicate_uri: str) -> List[BaseOutputPredicate]:
        pass

    def o(self, predicate_uri: str, target_uri: str) -> 'BaseOutputClass':
        pass

    def __init__(self, classes: List['ArrayClass']):
        self.classes = classes

    def iter_records(self) -> Iterable[ArrayRecord]:
        for cls in self.classes:
            for r in cls.iter_records():
                yield r

    def filter(self, conditions) -> 'LstArrayClass':
        new_classes = []
        for cls in self.classes:
            cls = cls.filter(conditions)
            if isinstance(cls, SubsetArrayClass) and len(cls.ids) == 0:
                continue
            new_classes.append(cls)
        return LstArrayClass(new_classes)

    def group_by(self, predicate: str) -> Iterable[Tuple[Any, 'ArrayClass']]:
        pass

    def get_record_by_id(self, rid: RecordID):
        pass





from dataclasses import dataclass

import numpy as np
from typing import Union, List, Tuple, Optional

from drepr.models import Path


@dataclass
class NoData:
    value: any


class ArrayAttr:
    def __init__(self, id: str, values: np.ndarray, path: Path, step2dim: List[Optional[int]], nodata: Optional[NoData]):
        self.id = id
        self.values = values
        self.values_shp = values.shape[:sum(int(d is not None) for d in step2dim)]
        self.steps = path.steps
        self.step2dim = step2dim
        self.nodata = nodata

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.values.shape

    @property
    def size(self) -> int:
        return self.values.size

    def get_data(self):
        if not isinstance(self.values, np.ndarray):
            self.values = np.asarray(self.values)
        return self.values

    def get_value(self, index: Tuple[int, ...]):
        return self.values[index]

    def set_value(self, index: Tuple[int, ...], val):
        self.values[index] = val


class ScalarAttr:
    def __init__(self, id: str, value):
        self.id = id
        self.value = value
        self.values = None
        self.nodata = None

    @property
    def shape(self) -> Tuple[int, ...]:
        return ()

    @property
    def size(self):
        return 1

    def get_data(self):
        if self.values is None:
            self.values = np.asarray([self.value])
        return self.values

    def get_value(self, _index: Tuple[int, ...]):
        return self.value

    def set_value(self, _index: Tuple[int, ...], val):
        self.value = val


Attribute = Union[ArrayAttr, ScalarAttr]



from typing import List, Type, Union

from drepr.utils.validator import Validator
from .path_parser import PathParser
from ..preprocessing import Preprocessing, PreprocessingType, PMap, PFilter, RMap, PSplit
from ..resource import Resource
from ..path import Path


class PreprocessingParser:
    """
    Preprocessing are defined as a list

    ```
    - type: <preprocessing_type>
      # other properties
    ```

    1. If <preprocessing_type> is `pmap`, its other properties are:

        ```
        - type: pmap
          [resource_id]: <resource_id>
          path: <path>
          [output]: <resource_id> (default is None)
          [change_structure]: null|true|false (default is null)
          code: str
        ```

    2. If <preprocessing_type> is `pfilter`, its properties are:

        ```
        - type: pfilter
          [resource_id]: <resource_id>
          path: <path>
          [output]: <resource_id> (default is None)
          code: str
        ```

    3. If <preprocessing_type> is `rmap`, its properties are:

        ```
        - type: rmap
          resource_id: <resource_id>
          path: <path>
          func_id: <func_id>
          [output]: <resource_id> (default is None)
        ```
    """
    PREPRO_TYPES = {x.value for x in PreprocessingType}

    def __init__(self, path_parser: PathParser):
        self.path_parser = path_parser

    def parse(self, default_resource_id: str, resources: List[Resource], conf: list) -> List[Preprocessing]:
        Validator.must_be_list(conf, "Parsing preprocessing")
        result = []

        for i, prepro in enumerate(conf):
            trace0 = f"Parsing preprocessing at position {i}"
            Validator.must_be_dict(prepro, trace0)
            Validator.must_have(prepro, "type", trace0)
            Validator.must_in(prepro['type'], self.PREPRO_TYPES, f"{trace0}\nParsing property `type`")
            prepro_type = PreprocessingType(prepro['type'])

            if "resource_id" in prepro:
                Validator.must_be_str(prepro["resource_id"], f"{trace0}\nParsing property `resource_id`")
                resource_id = prepro['resource_id']
            else:
                resource_id = default_resource_id

            trace1 = f"{trace0}\nParsing property `path`"
            Validator.must_have(prepro, "path", trace1)
            path = self.path_parser.parse(
                self.path_parser.get_resource(resources, resource_id, trace0),
                prepro['path'], trace1)

            if prepro_type == PreprocessingType.pmap:
                value = self.parse_pmap(resource_id, path, prepro, trace0)
            elif prepro_type == PreprocessingType.pfilter:
                value = self.parse_pfilter_psplit(resource_id, path, prepro, trace0, PFilter)
            elif prepro_type == PreprocessingType.psplit:
                value = self.parse_pfilter_psplit(resource_id, path, prepro, trace0, PSplit)
            elif prepro_type == PreprocessingType.rmap:
                value = self.parse_rmap(resource_id, path, prepro, trace0)
            else:
                raise NotImplemented(f"Not implement the parser for preprocessing function with type {prepro_type}")

            result.append(Preprocessing(prepro_type, value))
        return result

    def parse_pmap(self, resource_id: str, path: Path, prepro: dict, trace0: str) -> PMap:
        trace1 = f"{trace0}\nParsing property `code`"
        Validator.must_have(prepro, "code", trace1)
        Validator.must_be_str(prepro["code"], trace1)
        code = prepro["code"]

        if "output" in prepro and prepro['output'] is not None:
            trace1 = f"{trace0}\nParsing property `output`"
            Validator.must_be_str(prepro["output"], trace1)
            output = prepro["output"]
        else:
            output = None

        if "change_structure" in prepro and prepro["change_structure"] is not None:
            trace1 = f"{trace0}\nParsing property `change_structure`"
            Validator.must_be_bool(prepro["change_structure"], trace1)
            change_structure = prepro["change_structure"]
        else:
            change_structure = None

        return PMap(resource_id, path, code, output, change_structure)

    def parse_pfilter_psplit(self, resource_id: str, path: Path, prepro: dict, trace0: str, cls: Union[Type[PFilter], Type[PSplit]]) -> PFilter:
        trace1 = f"{trace0}\nParsing property `code`"
        Validator.must_have(prepro, "code", trace1)
        Validator.must_be_str(prepro["code"], trace1)
        code = prepro["code"]

        if "output" in prepro:
            trace1 = f"{trace0}\nParsing property `output`"
            Validator.must_be_str(prepro["output"], trace1)
            output = prepro["output"]
        else:
            output = None

        return cls(resource_id, path, code, output)

    def parse_rmap(self, resource_id: str, path: Path, prepro: dict, trace0: str) -> RMap:
        trace1 = f"{trace0}\nParsing property `func_id`"
        Validator.must_have(prepro, "func_id", trace1)
        Validator.must_be_str(prepro["func_id"], trace1)
        func_id = prepro["func_id"]

        if "output" in prepro:
            trace1 = f"{trace0}\nParsing property `output`"
            Validator.must_be_str(prepro["output"], trace1)
            output = prepro["output"]
        else:
            output = None

        return RMap(resource_id, path, func_id, output)



from typing import List, Dict, Tuple, Callable, Any, Optional, TYPE_CHECKING
import numpy as np
from drepr.models import SemanticModel, DRepr, Alignment, RangeAlignment, Attr, Edge, ClassNode
from drepr.outputs.array_backend.array_attr import NoData
from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_output_class import BaseOutputClass
from drepr.outputs.graph_backend.lst_graph_class import LstGraphClass

if TYPE_CHECKING:
    from drepr.outputs import GraphBackend
    from drepr.outputs.graph_backend.graph_class import GraphClass
from drepr.outputs.prop_data_ndarray import PropDataNDArray, IndexPropRange


class GraphPredicate(BaseOutputPredicate):

    def __init__(self, backend: 'GraphBackend', cls: 'GraphClass', edges: List[Edge]) -> None:
        super().__init__()
        self.sm = backend.sm
        self.cls = cls
        self.attrs = backend.drepr.attrs
        self.uri = edges[0].label
        self.edges = edges
        self.targets: Optional[LstGraphClass] = None

    def _init(self, backend: 'GraphBackend'):
        if len(self.edges) > 0:
            self.targets = LstGraphClass([
                backend.cid(edge.target_id)
                for edge in self.edges
                if isinstance(backend.sm.nodes[edge.target_id], ClassNode)
            ])
            assert len(self.targets) == len(self.edges)

    def ndarray_size(self) -> int:
        return len(self.cls)

    def as_ndarray(self, indexed_predicates: List[BaseOutputPredicate]) -> PropDataNDArray:
        if len(self.edges) > 1 or any(len(p.edges) > 1 for p in indexed_predicates):
            raise Exception("Cannot convert values of this predicate to an ndarray indexed by other predicates "
                            "because values for one entry may be greater than one")

        attr_id = self.sm.nodes[self.edges[0].target_id].attr_id
        nodata = None
        for attr in self.attrs:
            if attr.id == attr_id:
                if len(attr.missing_values) > 0:
                    # the engine already converted all missing values to None
                    nodata = NoData(None)
                break

        # cannot handle the case where uri of one of the index predicate is duplicated, since we cannot retrieve correct
        # value using graph backend, if we want to support it, we need to modify the backend to keep the null value as None
        for p in indexed_predicates:
            assert len(p.edges) == 1

        n_records = len(self.cls)
        if n_records == 0:
            data = np.asarray([])
            index_props_range = [IndexPropRange(0, 0) for _ in indexed_predicates]
            index_props = [np.asarray([]) for p in indexed_predicates]
        else:
            record = next(self.cls.iter_records())
            data = np.fromiter((r.s(self.uri) for r in self.cls.iter_records()), type(record.s(self.uri)), count=n_records)
            index_props_range = [IndexPropRange(0, 0) for _ in indexed_predicates]
            index_props = [
                np.fromiter((r.s(p.uri) for r in self.cls.iter_records()), type(record.s(p.uri)), count=n_records)
                for p in indexed_predicates
            ]
        return PropDataNDArray(data, nodata, index_props_range, index_props)

    def o(self) -> Optional[LstGraphClass]:
        return self.targets




from collections import defaultdict
from dataclasses import dataclass
from typing import List, Dict, Tuple, Callable, Any, Optional, Union

from drepr.models import SemanticModel, Edge, ClassNode, DataNode


class Namespace:

    def __init__(self, base: str, seen_uris: Optional[Dict[str, str]] = None):
        self._base = base
        self.seen_uris = seen_uris or {}

    def __getattr__(self, item):
        if item not in self.seen_uris:
            uri = self._base + item
            self.seen_uris[item] = uri
        return self.seen_uris[item]

    def __getitem__(self, item):
        if item not in self.seen_uris:
            uri = self._base + item
            self.seen_uris[item] = uri
        return self.seen_uris[item]


class DataProp:
    def __init__(self, sm: SemanticModel, edge: Edge):
        self.sm = sm
        self.edge = edge
        self.data_id = edge.target_id

    @property
    def label(self):
        return self.edge.label

    @property
    def edge_id(self):
        return self.edge.edge_id


class ObjectProp:
    def __init__(self, indexed_sm: 'IndexedSM', edge: Edge, object: 'SMClass'):
        self.indexed_sm = indexed_sm
        self.edge = edge
        self.object: SMClass = object

    @property
    def label(self):
        return self.edge.label

    @property
    def edge_id(self):
        return self.edge.edge_id


class SMClass:
    def __init__(self, indexed_sm: 'IndexedSM', node_id: str):
        self.indexed_sm = indexed_sm
        self.sm = indexed_sm.sm
        self.node_id = node_id
        self.pk_attr: str = ""
        self.uri_attr: Optional[str] = None
        self.predicates: Dict[str, List[Union[DataProp, ObjectProp]]] = defaultdict(lambda: [])
        self.index_po: Dict[Tuple[str, str], List[SMClass]] = defaultdict(lambda: [])

    def _init(self):
        for e in self.sm.iter_outgoing_edges(self.node_id):
            if isinstance(self.sm.nodes[e.target_id], ClassNode):
                o_uri = self.sm.nodes[e.target_id].label
                o = self.indexed_sm.get_class_by_id(e.target_id)
                self.index_po[(e.label, o_uri)].append(o)
                self.predicates[e.label].append(ObjectProp(self.indexed_sm, e, o))
            else:
                if e.label == 'drepr:uri':
                    self.uri_attr = e.target_id
                else:
                    self.predicates[e.label].append(DataProp(self.sm, e))

            if e.is_subject:
                # this requires us to analyze the d-repr output first.
                self.pk_attr = e.target_id

    def is_blank(self) -> bool:
        return self.uri_attr is None

    @property
    def label(self):
        return self.sm.nodes[self.node_id].label

    def p(self, predicate_uri: str) -> List[Union[DataProp, ObjectProp]]:
        """Get list of predicates by their uri"""
        return self.predicates[predicate_uri]

    def o(self, predicate_uri: str, target_uri: str) -> List['SMClass']:
        """Get list of target classes"""
        return self.index_po[(predicate_uri, target_uri)]


class IndexedSM:
    def __init__(self, sm: SemanticModel):
        self.sm = sm
        self.sm_classes: Dict[str, List[SMClass]] = defaultdict(lambda: [])
        self.id2cls: Dict[str, SMClass] = {}

        for n in sm.iter_class_nodes():
            cls = SMClass(self, n.node_id)
            self.sm_classes[n.label].append(cls)
            self.id2cls[n.node_id] = cls

        for c in self.id2cls.values():
            c._init()

    def c(self, class_uri: str) -> List[SMClass]:
        """
        Get a list of classes defined in the semantic model by the URI.
        """
        return self.sm_classes[class_uri]

    def get_class_by_id(self, node_id: str) -> SMClass:
        return self.id2cls[node_id]

    # noinspection PyMethodMayBeStatic
    def ns(self, uri: str) -> Namespace:
        return Namespace(uri)



import enum
from abc import ABC, abstractmethod
from typing import Dict, Union, List, Iterable, Any, Optional, Callable

from drepr import DRepr
from drepr.models import SemanticModel
from drepr.outputs.record_id import RecordID
from drepr.outputs.base_lst_output_class import BaseLstOutputClass
from drepr.outputs.base_output_class import BaseOutputClass
from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_record import BaseRecord
from drepr.outputs.namespace import Namespace, PrefixedNamespace


class BaseOutputSM(ABC):

    @classmethod
    @abstractmethod
    def from_drepr(cls, ds_model: Union[DRepr, str], resources: Union[str, Dict[str, str]], inject_class_id: Callable[[str], str]=None) -> "BaseOutputSM":
        """
        :param ds_model:
        :param resources:
        :param inject_class_id: this optional function allows users to inject more information into the class_id
        :return:
        """
        pass

    @abstractmethod
    def iter_classes(self) -> Iterable[BaseOutputClass]:
        pass

    @abstractmethod
    def get_record_by_id(self, rid: RecordID) -> BaseRecord:
        pass

    @abstractmethod
    def c(self, class_uri: str) -> BaseLstOutputClass:
        """Get list of classes based on their URIs"""
        pass

    @abstractmethod
    def cid(self, class_id: str) -> BaseOutputClass:
        """Get a class based on its id in the semantic model"""
        pass

    @abstractmethod
    def get_sm(self) -> SemanticModel:
        pass

    def ns(self, uri: str) -> Namespace:
        sm = self.get_sm()
        if not hasattr(sm, '_inverted_prefixes'):
            sm._inverted_prefixes = {v: k for k, v in sm.prefixes.items()}

        if uri in sm._inverted_prefixes:
            return PrefixedNamespace(uri, sm._inverted_prefixes[uri])
        return Namespace(uri)


class FCondition:
    def __init__(self, predicate_uri: str, op: str, val: Optional[Any]=None):
        self.predicate_uri = predicate_uri
        self.op = FConditionOp(op)
        self.val = val


class FConditionOp(enum.Enum):
    eq = "=="
    gt = ">"
    gte = ">="
    lt = "<"
    lte = "<="
    is_in = "is_in"


def identity(x):
    return x




import re

from drepr.utils.validator import Validator, InputError
from ..sm import SemanticModel, ClassNode, DataNode, Edge, LiteralNode, DataType


class SMParser:
    """
    SM has the following schema

    ```
    semantic_model:
      data_nodes:
        <attr_id>: <class_id>--<predicate>[^^<semantic type>]
        # other attributes
      relations:
        - <source_class_id>--<predicate>--<target_class_id>
        # other relations
      literal_nodes:
        - <source_class_id>--<predicate>--<value>
      subjects:
        <class_id>: <attr_id>
      prefixes:
        <prefix>: <uri>
    ```
    """
    SM_KEYS = {"data_nodes", "subjects", "literal_nodes", "prefixes", "relations"}
    DATA_TYPE_VALUES = {x.value for x in DataType}

    REG_SM_CLASS = re.compile(r"^((.+):\d+)$")
    REG_SM_DNODE = re.compile(r"^((?:(?!--).)+:\d+)--((?:(?!\^\^).)+)(?:\^\^(.+))?$")
    REG_SM_LNODE = re.compile(
        r"^((?:(?!--).)+:\d+)--((?:(?!--).)+)--((?:(?!\^\^).)+)(?:\^\^(.+))?$")
    REG_SM_REL = re.compile(r"^((?:(?!--).)+:\d+)--((?:(?!--).)+)--((?:(?!--).)+:\d+)$")

    @classmethod
    def parse(cls, sm: dict) -> SemanticModel:
        Validator.must_be_subset(cls.SM_KEYS, sm.keys(), "properties of semantic model", "Parsing the semantic model")

        Validator.must_have(sm, "data_nodes", "Parsing the semantic model")
        trace0 = "Parsing `data_nodes` of the semantic model"
        Validator.must_be_dict(sm['data_nodes'], trace0)

        nodes = {}
        edges = {}

        for attr_id, stype in sm['data_nodes'].items():
            trace1 = f"{trace0}\nParsing data node `{attr_id}`"
            m = cls.REG_SM_DNODE.match(stype)
            if m is None:
                raise InputError(f"{trace1}\nERROR: the value of data node does not match with the format")

            # do something with the data node
            class_id = m.group(1)
            class_name = cls.REG_SM_CLASS.match(m.group(1)).group(2)
            predicate = m.group(2)
            data_type = m.group(3)

            if data_type is not None:
                Validator.must_in(data_type, cls.DATA_TYPE_VALUES, f"{trace1}\nParsing data type")
                data_type = DataType(data_type)
            else:
                data_type = None

            if class_id not in nodes:
                nodes[class_id] = ClassNode(node_id=class_id, label=class_name)

            data_node = DataNode(node_id=f"dnode:{attr_id}", attr_id=attr_id, data_type=data_type)
            nodes[data_node.node_id] = data_node
            edges[len(edges)] = Edge(len(edges), class_id, data_node.node_id, predicate)

        if 'relations' in sm:
            trace0 = f"Parsing `relations` of the semantic model"
            Validator.must_be_list(sm['relations'], trace0)
            for i, node in enumerate(sm['relations']):
                trace1 = f"{trace0}\nParsing relation at position {i}: {node}"
                Validator.must_be_str(node, trace1)
                m = cls.REG_SM_REL.match(node)
                if m is None:
                    raise InputError(f"{trace1}\nERROR: value of the relation does not match with the format")

                edges[len(edges)] = Edge(len(edges), source_id=m.group(1), target_id=m.group(3), label=m.group(2))

        if 'literal_nodes' in sm:
            trace0 = f"Parsing `literal_nodes` of the semantic model"
            Validator.must_be_list(sm['literal_nodes'], trace0)
            for i, node in enumerate(sm['literal_nodes']):
                trace1 = f"{trace0}\nParsing literal node at position {i}: {node}"
                Validator.must_be_str(node, trace1)
                m = cls.REG_SM_LNODE.match(node)
                if m is None:
                    raise InputError(f"{trace1}\nERROR: value of the literal node does not match with the format")

                class_id = m.group(1)
                class_name = cls.REG_SM_CLASS.match(m.group(1)).group(2)
                predicate = m.group(2)

                data_type = m.group(4)
                if data_type is not None:
                    Validator.must_in(data_type, cls.DATA_TYPE_VALUES, f"{trace1}\nParsing data type")
                data_type = DataType(data_type)

                if class_id not in nodes:
                    nodes[class_id] = ClassNode(node_id=class_id, label=class_name)

                literal_node = LiteralNode(
                    node_id=f"lnode:{i}",
                    value=m.group(3),
                    data_type=data_type)
                nodes[literal_node.node_id] = literal_node
                edges[len(edges)] = Edge(len(edges), source_id=class_id, target_id=literal_node.node_id, label=predicate)

        if 'prefixes' in sm:
            trace0 = f"Parsing `prefixes` of the semantic model"
            Validator.must_be_dict(sm['prefixes'], trace0)
            for prefix, uri in sm['prefixes'].items():
                Validator.must_be_str(uri, f"{trace0}\nParse prefix {prefix}")

            prefixes = dict(sm['prefixes'])
        else:
            prefixes = {}

        if 'subjects' in sm:
            trace0 = f"Parsing `subjects` of the semantic model"
            Validator.must_be_dict(sm['subjects'], trace0)
            for class_id, attr_id in sm['subjects'].items():
                Validator.must_be_str(attr_id, f"{trace0}\nParsing subject of class {class_id}")
                target_id = f"dnode:{attr_id}"
                for edge in edges.values():
                    if edge.source_id == class_id and edge.target_id == target_id:
                        edge.is_subject = True
                        break

        return SemanticModel(nodes, edges, prefixes)



from dataclasses import dataclass
from enum import Enum
from typing import List, Union, Optional

from .path import Path


class Sorted(Enum):
    Null = "none"
    Ascending = "ascending"
    Descending = "descending"


class ValueType(Enum):
    Unspecified = "unspecified"
    Int = "int"
    Float = "float"
    String = "str"
    List_Int = "list[int]"
    List_Float = "list[float]"
    List_String = "list[str]"


@dataclass
class Attr:
    id: str
    resource_id: str
    path: Path
    missing_values: List[Optional[Union[str, int, float]]]
    unique: bool = False
    sorted: Sorted = Sorted.Null
    value_type: ValueType = ValueType.Unspecified

    @staticmethod
    def deserialize(raw: dict) -> "Attr":
        attr = Attr(**raw)
        attr.path = Path.deserialize(raw['path'])
        attr.sorted = Sorted(attr.sorted)
        attr.value_type = ValueType(attr.value_type)

        return attr



from drepr.models.parse_v2.path_parser import PathParserV2
from drepr.utils.validator import *

from ..parse_v1.align_parser import AlignParser
from ..parse_v1.attr_parser import AttrParser
from ..parse_v1.preprocessing_parser import PreprocessingParser
from ..parse_v1.resource_parser import ResourceParser

from .sm_parser import SMParser
from ..sm import SemanticModel


class ReprV2Parser:
    """
    The D-REPR language version 2 has similar to the schema of the first version.

    Difference with previous features:
    1. For spreadsheet columns, they can the letter instead of number
    2. Semantic model configuration is changed to focus on classes
    """

    TOP_KEYWORDS = {
        "version", "resources", "preprocessing", "attributes", "alignments", "semantic_model"
    }
    DEFAULT_RESOURCE_ID = "default"

    @classmethod
    def parse(cls, raw: dict):
        from ..drepr import DRepr

        Validator.must_be_subset(cls.TOP_KEYWORDS,
                                 raw.keys(),
                                 setname="Keys of D-REPR configuration",
                                 error_msg="Parsing D-REPR configuration")

        for prop in ['version', 'resources', 'attributes']:
            Validator.must_have(raw, prop, error_msg="Parsing D-REPR configuration")

        Validator.must_equal(raw['version'], '2', "Parsing D-REPR configuration version")
        resources = ResourceParser.parse(raw['resources'])

        if len(resources) == 1:
            default_resource_id = resources[0].id
        else:
            default_resource_id = ResourceParser.DEFAULT_RESOURCE_ID

        path_parser = PathParserV2()
        preprocessing = PreprocessingParser(path_parser).parse(default_resource_id, resources,
                                                               raw.get('preprocessing', []))
        attrs = AttrParser(path_parser).parse(default_resource_id, resources, raw['attributes'])
        aligns = AlignParser.parse(raw.get('alignments', []))

        if 'semantic_model' in raw:
            sm = SMParser.parse(raw['semantic_model'])
            sm.prefixes.update(SemanticModel.get_default_prefixes())
        else:
            sm = None

        return DRepr(resources, preprocessing, attrs, aligns, sm)



import numpy as np, fiona
from typing import List, Dict, Tuple, Callable, Any, Optional, Union

from drepr.executors.readers.ra_reader import NDArrayReader, Index
from drepr.models import IndexExpr, RangeExpr


class ShapefileReader(NDArrayReader):
    """
    Reader for shapefile resources
    """
    def __init__(self, filename: str, features: List[dict]):
        self.metadata = {"filename": filename}
        self.data = features

    @staticmethod
    def from_file(resource_file):
        with fiona.open(resource_file, 'r') as f:
            features = list(f)
        return ShapefileReader(resource_file, features)

    def len(self) -> int:
        # count the global metadata at '@'
        return len(self.data) + 1

    def len_range(self) -> int:
        return len(self.data)

    def get_value(self, index: List[Index]):
        if index[0] == '@':
            # read metadata
            ptr = self.metadata
            for i in range(1, len(index)):
                ptr = ptr[index[i]]
            return ptr

        ptr = self.data
        for idx in index:
            ptr = ptr[idx]
        return ptr

    def set_value(self, index: List[Index], value):
        if index[0] == '@':
            # update metadata
            ptr = self.metadata
            for i in range(1, len(index) - 1):
                ptr = ptr[index[i]]
            ptr[index[-1]] = value
            return

        ptr = self.data[index[0]]
        for i in range(1, len(index) - 1):
            ptr = ptr[index[i]]
        ptr[index[-1]] = value

    def select(self, steps: List[Union[IndexExpr, RangeExpr]]):
        # steps.0 must be index because there is no structure change
        if isinstance(steps[0], IndexExpr) and steps[0].val == '@':
            # select metadata
            value = self.metadata
            for i in range(1, len(steps)):
                assert isinstance(steps[i], IndexExpr)
                value = value[steps[i].val]
            return value

        value = self._recursive_select(self.data, steps, 0)
        if isinstance(value, list):
            return np.asarray(value)
        return value

    def _recursive_select(self, value, steps: List[Union[IndexExpr, RangeExpr]], i):
        if i >= len(steps):
            return value

        if isinstance(steps[i], IndexExpr):
            return self._recursive_select(value[steps[i].val], steps, i+1)

        if isinstance(steps[i], RangeExpr):
            # special case for 2D geometry
            if all(isinstance(steps[j], RangeExpr) and steps[j].is_select_all() for j in range(i, len(steps))):
                return value

            if steps[i].end is None:
                end = len(value)
            else:
                end = steps[i].end
            return [self._recursive_select(value[j], steps, i+1)
                    for j in range(steps[i].start, end, steps[i].step)]

        raise ValueError("Invalid step: %s" % steps[i])



from typing import List, Dict, Tuple, Callable, Any, Optional, Iterable, Union, TYPE_CHECKING

from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_record import BaseRecord
from drepr.outputs.base_lst_output_class import BaseLstOutputClass
from drepr.outputs.base_output_class import BaseOutputClass

if TYPE_CHECKING:
    from drepr.outputs.graph_backend.graph_class import GraphClass
from drepr.outputs.record_id import GraphRecordID


class LstGraphClass(BaseLstOutputClass):

    def __init__(self, classes: List['GraphClass']) -> None:
        super().__init__()
        self.classes = classes

    def iter_records(self) -> Iterable[BaseRecord]:
        pass

    def get_record_by_id(self, rid: GraphRecordID) -> BaseRecord:
        pass

    def p(self, predicate_uri: str) -> List[BaseOutputPredicate]:
        pass

    def o(self, predicate_uri: str, target_uri: str) -> List['BaseOutputClass']:
        pass

    def filter(self, conditions: Union['FCondition', List['FCondition']]) -> 'BaseSubsetOutputClass':
        pass

    def group_by(self, predicate: Union[str, BaseOutputPredicate]) -> Iterable[Tuple[Any, 'BaseSubsetOutputClass']]:
        pass

    def __iter__(self) -> Iterable[BaseOutputClass]:
        return iter(self.classes)

    def __len__(self) -> int:
        return len(self.classes)

    def __getitem__(self, item: int) -> BaseOutputClass:
        return self.classes[item]


from typing import Dict

from drepr.executors.readers.ra_reader import NDArrayReader


class Context:
    def __init__(self, reader: NDArrayReader):
        self.reader = reader

    def get_value(self, index):
        return self.reader.get_value(index)


from typing import List, Dict, Tuple, Callable, Any, Optional

from drepr.executors.readers.ra_reader import NDArrayReader


class ReaderContainer:
    """
    This contains the list of ndarray reader that can be injected by the users so that they can extend the
    coverage of the system
    """
    instance = None

    def __init__(self):
        self.readers = {}

    @staticmethod
    def get_instance():
        if ReaderContainer.instance is None:
            ReaderContainer.instance = ReaderContainer()
        return ReaderContainer.instance

    def set(self, reader_id: str, reader: NDArrayReader):
        self.readers[reader_id] = reader

    def get(self, reader_id: str):
        return self.readers[reader_id]

    def delete(self, reader_id: str):
        if reader_id in self.readers:
            del self.readers[reader_id]

